/* using type list to order the singletons' destruction */
#ifndef __SINGLETON__
#define __SINGLETON__
#include "typelist.h"
#include <cstddef>
#include <vector>
#ifndef _MSC_VER
using std::size_t;
#endif
using std::vector;
typedef void (*__destroy_func__) ();
template <typename _Type> class Singleton;

template <typename _TypeList>
struct _type_list_destructor_register
{
	static inline void Travel (__destroy_func__ df)
	{
		typedef _TypeList::this_type this_type;
		typedef _TypeList::this_list this_list;
		Singleton<this_type>::instance_manager::get().register_destructor(df);
		_type_list_destructor_register<this_list>::Travel (df);
	}
};

template <>
struct _type_list_destructor_register <end_of_type_list>
{ static inline void Travel (__destroy_func__) {} };


template <typename _Type>
class Singleton
{
public:
	typedef _Type type;
	typedef typename _Type::destroy_dependents dependents;
	typedef Singleton<_Type> singleton;

	class instance_manager;
	friend instance_manager;

	class instance_manager
	{
		friend singleton;

	public:
		struct _creator
		{
			_creator () { instance_manager::get(); }
			void nop () const {}
		};
		static _creator creator;

	public:
		~instance_manager () { destroy_instance (); }
		void register_destructor (__destroy_func__ destructor) 
		{ _M_managed.push_back(destructor); }
		bool check_instance () { return _M_instance ? true : false; }
		type *instance () 
		{
			return _M_instance ? 
			_M_instance : 
			_M_instance = singleton::_create_instance(); 
		}

		void destroy_instance ()
		{
			for (size_t i=0; i<_M_managed.size(); i++)	
				(*_M_managed[i])(); 
			singleton::_destroy_instance (_M_instance);
			_M_instance = 0;
		}

		static instance_manager &get ()
		{
			static instance_manager im;
			creator.nop ();
			return im;
		}

		static inline void destroy ()
		{ singleton::instance_manager::get().destroy_instance(); }

	private:
		instance_manager () : _M_instance(0)
		{ _type_list_destructor_register<dependents>::Travel(destroy); }
		type *_M_instance;
		vector<__destroy_func__> _M_managed;
	};

	static inline type & instance () 
	{ return *instance_manager::get().instance(); }

private:
	static inline type *_create_instance () { return new type; }
	static inline void _destroy_instance (type *p) { delete p; }
};

#define DECLARE_SINGLETON_DEPENDENTS(type,typelist) \
typedef typelist destroy_dependents; \
friend Singleton< type >;

#define DECLARE_SINGLETON_NO_DEPENDENTS(type) \
typedef end_of_type_list destroy_dependents; \
friend Singleton< type >;

#define IMPLEMENT_SINGLETON(type) \
Singleton< type >::instance_manager::\
_creator Singleton< type >::instance_manager::creator;

#define GET_SINGLETON( type ) \
(Singleton< type >::instance())

template <typename _Type>
typename Singleton<_Type>::instance_manager::_creator 
Singleton<_Type>::instance_manager::creator;

#endif//__SINGLETON__
